@@ -1,4 +1,4 @@
-"""Custom exceptions for paraDime.
+"""Custom exceptions for ParaDime.
 """
 
 
@@ -14,5 +14,17 @@ class NoDatasetRegisteredError(Exception):
     """Attempted to access the dataset before registration."""
 
 
+class LossNotDeterminedError(Exception):
+    """Attempted to access loss before it was determined."""
+
+
 class UnsupportedConfigurationError(Exception):
     """Combination of parameters and/or objects not supported."""
+
+
+class SpecificationError(Exception):
+    """Specification not supported."""
+
+
+class CUDANotAvailableError(Exception):
+    """CUDA not available."""
