@@ -1,4 +1,4 @@
-"""Utility functions for paraDime.
+"""Utility functions for ParaDime.
 
 The :mod:`paradime.utils` subpackage includes various modules that implement
 utility functions for logging, plotting, and input conversion.
@@ -6,6 +6,7 @@ utility functions for logging, plotting, and input conversion.
 
 from paradime.utils import convert
 from paradime.utils import logging
+from paradime.utils import parsing
 from paradime.utils import plotting
 from paradime.utils import seed
 
